from .utils import (
    get_graph_id,
    get_graph_from_id,
    opt_angles_for_graph,
    angles_to_qaoa_format,
    angles_to_qiskit_format,
    angles_to_qtensor_format,
    angles_from_qiskit_format,
    beta_to_qaoa_format,
    gamma_to_qaoa_format,
    get_fixed_angles,
    get_full_qaoa_dataset_table_row,
    get_full_qaoa_dataset_table,
    get_3_reg_dataset_table,
    get_3_reg_dataset_table_row,
    qaoa_maxcut_energy,
)



import networkx as nx
import numpy as np
import pandas as pd
from qiskit import QuantumRegister, ClassicalRegister, execute
from qiskit.providers.aer import AerSimulator
from functools import partial
from pathlib import Path
import copy
import pytest

from qiskit.quantum_info import Statevector

from QAOAKit import (
    opt_angles_for_graph,
    get_fixed_angles,
    get_graph_id,
    get_graph_from_id,
    angles_to_qaoa_format,
    beta_to_qaoa_format,
    gamma_to_qaoa_format,
    angles_to_qiskit_format,
    angles_to_qtensor_format,
    get_3_reg_dataset_table,
    get_3_reg_dataset_table_row,
    get_full_qaoa_dataset_table_row,
    get_full_qaoa_dataset_table,
    qaoa_maxcut_energy,
    angles_from_qiskit_format,
)
from QAOAKit.utils import (
    obj_from_statevector,
    maxcut_obj,
    isomorphic,
    load_weights_into_dataframe,
    load_weighted_results_into_dataframe,
    get_adjacency_matrix,
)
from QAOAKit.qaoa import get_maxcut_qaoa_circuit, get_maxcut_qaoa_qiskit_circuit
from QAOAKit.examples_utils import get_20_node_erdos_renyi_graphs

from qiskit_optimization import QuadraticProgram
from qiskit.algorithms.minimum_eigen_solvers.qaoa import QAOAAnsatz

test_utils_folder = Path(__file__).parent


def test_retrieval():
    for n in range(3, 10):
        G = nx.complete_graph(n)
        for p in range(1, 4):
            angles = opt_angles_for_graph(G, p)
            assert len(angles["beta"]) == p
            assert len(angles["gamma"]) == p
            assert isinstance(
                get_full_qaoa_dataset_table_row(G, p)[
                    ["C_opt", "graph_id", "C_{true opt}"]
                ],
                pd.Series,
            )


def test_fixed_angle_retrieval():
    for d in range(3, 5):
        for p in range(1, 4):
            angles = get_fixed_angles(d, p)
            assert type(angles) is dict
            assert len(angles["beta"]) == p
            assert len(angles["gamma"]) == p


def test_tables_consistency():
    p = 1
    G1 = nx.complete_graph(5)
    row = get_full_qaoa_dataset_table_row(G1, p)
    G2 = get_graph_from_id(row["graph_id"], 5)
    assert isomorphic(G1, G2)


def test_angle_conversion():
    full_qaoa_dataset_table = get_full_qaoa_dataset_table()
    for n_qubits in [3, 5]:
        for p in [2, 3]:
            df = full_qaoa_dataset_table.reset_index()
            df = df[(df["n"] == n_qubits) & (df["p_max"] == p)]
            for _, row in df.iterrows():
                obj = partial(maxcut_obj, w=get_adjacency_matrix(row["G"]))
                opt_cut = row["C_opt"]
                angles = angles_to_qaoa_format(
                    opt_angles_for_graph(row["G"], row["p_max"])
                )
                assert np.allclose(angles["beta"], beta_to_qaoa_format(row["beta"]))
                assert np.allclose(angles["gamma"], gamma_to_qaoa_format(row["gamma"]))
                qc = get_maxcut_qaoa_circuit(row["G"], angles["beta"], angles["gamma"])
                backend = AerSimulator(method="statevector")
                res = backend.run(qc).result()
                sv = res.get_statevector()
                obj_val = obj_from_statevector(sv, obj)
                assert np.isclose(opt_cut, obj_val)


def test_qaoa_maxcut_energy():
    full_qaoa_dataset_table = get_full_qaoa_dataset_table()
    for n_qubits in [3, 4]:
        p = 3
        df = full_qaoa_dataset_table.reset_index()
        df = df[(df["n"] == n_qubits) & (df["p_max"] == p)]
        for _, row in df.iterrows():
            assert np.isclose(
                row["C_opt"],
                qaoa_maxcut_energy(
                    row["G"],
                    beta_to_qaoa_format(row["beta"]),
                    gamma_to_qaoa_format(row["gamma"]),
                ),
            )


def test_qiskit_angle_conversion():
    full_qaoa_dataset_table = get_full_qaoa_dataset_table()
    for n_qubits in [3, 4]:
        p = 3
        df = full_qaoa_dataset_table.reset_index()
        df = df[(df["n"] == n_qubits) & (df["p_max"] == p)]
        for _, row in df.iterrows():
            angles = opt_angles_for_graph(row["G"], row["p_max"])
            G = row["G"]
            qc, C, offset = get_maxcut_qaoa_qiskit_circuit(
                G, p, angles_to_qiskit_format(angles)
            )
            backend = AerSimulator(method="statevector")
            sv = Statevector(backend.run(qc).result().get_statevector())
            obj_val = -(sv.expectation_value(C) + offset)
            opt_cut = row["C_opt"]
            assert np.isclose(opt_cut, obj_val)


def test_qiskit_qaoa_circuit():
    full_qaoa_dataset_table = get_full_qaoa_dataset_table()
    for n_qubits in [3, 4]:
        p = 1
        df = full_qaoa_dataset_table.reset_index()
        df = df[(df["n"] == n_qubits) & (df["p_max"] == p)]
        for _, row in df.iterrows():
            angles1 = opt_angles_for_graph(row["G"], row["p_max"])
            G = row["G"]
            qc1, C, offset = get_maxcut_qaoa_qiskit_circuit(
                G, p, angles_to_qiskit_format(angles1)
            )
            backend = AerSimulator(method="statevector")
            sv1 = Statevector(backend.run(qc1).result().get_statevector())
            angles2 = angles_to_qaoa_format(
                opt_angles_for_graph(row["G"], row["p_max"])
            )
            qc2 = get_maxcut_qaoa_circuit(row["G"], angles2["beta"], angles2["gamma"])
            sv2 = Statevector(backend.run(qc2).result().get_statevector())
            assert sv1.equiv(sv2)


def test_weighted_qaoa():
    G = nx.petersen_graph()
    np.random.seed(42)
    for (u, v) in G.edges():
        G[u][v]["weight"] = np.random.uniform(low=1, high=5)
    assert nx.is_weighted(G)
    backend = AerSimulator(method="statevector")
    p = 3
    angles = {
        "beta": np.random.uniform(low=0, high=np.pi / 2, size=3),
        "gamma": np.random.uniform(low=0, high=np.pi, size=3),
    }
    qc1, C, offset = get_maxcut_qaoa_qiskit_circuit(
        G, p, angles_to_qiskit_format(angles)
    )
    sv1 = Statevector(backend.run(qc1).result().get_statevector())
    angles2 = angles_to_qaoa_format(angles)
    qc2 = get_maxcut_qaoa_circuit(G, angles2["beta"], angles2["gamma"])
    sv2 = Statevector(backend.run(qc2).result().get_statevector())
    assert sv1.equiv(sv2)


def test_load_weighted_results():
    p = 2
    n = 5

    folder_path = Path(test_utils_folder, f"../data/weighted_angle_dat/p={p}/")
    df_weights = load_weights_into_dataframe(folder_path)
    df = load_weighted_results_into_dataframe(folder_path, p, n, df_weights)
    assert np.all(
        np.isclose(
            df.head(100).apply(
                lambda row: qaoa_maxcut_energy(
                    row["G"],
                    beta_to_qaoa_format(row["beta"]),
                    gamma_to_qaoa_format(row["gamma"]),
                ),
                axis=1,
            ),
            df.head(100)["C_opt"],
        )
    )


def test_qtensor_angle_conversion():
    qtensor = pytest.importorskip("qtensor")
    full_qaoa_dataset_table = get_full_qaoa_dataset_table()
    for n_qubits in [3, 4]:
        p = 3
        df = full_qaoa_dataset_table.reset_index()
        df = df[(df["n"] == n_qubits) & (df["p_max"] == p)]
        for _, row in df.iterrows():
            angles = angles_to_qtensor_format(
                opt_angles_for_graph(row["G"], row["p_max"])
            )
            G = row["G"]
            obj_val = qtensor.QAOA_energy(G, angles["gamma"], angles["beta"])
            opt_cut = row["C_opt"]
            assert np.isclose(opt_cut, obj_val)


def test_from_qiskit_conversion():
    angles = {
        "beta": np.random.uniform(low=0, high=np.pi / 2, size=3),
        "gamma": np.random.uniform(low=0, high=np.pi, size=3),
    }
    angles2 = angles_from_qiskit_format(angles_to_qiskit_format(copy.deepcopy(angles)))
    assert np.all(np.isclose(angles["beta"], angles2["beta"]))
    assert np.all(np.isclose(angles["gamma"], angles2["gamma"]))


def test_3_reg_table():
    df = get_3_reg_dataset_table().reset_index()
    assert len(df) >= 1000

    # test that the angles are correct
    for _, row in df[(df["p_max"] == 1) | (df["p_max"] == 2)].head(50).iterrows():
        G = row["G"]
        angles = angles_to_qaoa_format({"beta": row["beta"], "gamma": row["gamma"]})
        assert np.isclose(
            qaoa_maxcut_energy(G, angles["beta"], angles["gamma"]), row["C_opt"]
        )

    # test that the optima match the full qaoa table
    for _, row in df[(df["p_max"] == 1) | (df["p_max"] == 2)].head(50).iterrows():
        G = row["G"]
        angles = angles_to_qaoa_format({"beta": row["beta"], "gamma": row["gamma"]})
        assert np.isclose(
            qaoa_maxcut_energy(G, angles["beta"], angles["gamma"]), row["C_opt"]
        )


def test_fixed_angles_3_reg():
    df = get_3_reg_dataset_table().sample(n=10).reset_index()

    for _, row in df.iterrows():
        angles = angles_to_qaoa_format(get_fixed_angles(3, row["p_max"]))
        assert np.isclose(
            qaoa_maxcut_energy(row["G"], angles["beta"], angles["gamma"]),
            row["C_fixed"],
            rtol=1e-4,
        )


def test_get_opt_angles_large_3_reg():
    for s in range(10):
        G = nx.random_regular_graph(3, 10, seed=s)
        for p in [1, 2]:
            angles = angles_to_qaoa_format(opt_angles_for_graph(G, p))
            row = get_3_reg_dataset_table_row(G, p)
            assert np.isclose(
                row["C_opt"], qaoa_maxcut_energy(G, angles["beta"], angles["gamma"])
            )


def test_example_in_README():
    # build graph
    G = nx.star_graph(5)
    # grab optimal angles
    p = 3
    angles = angles_to_qaoa_format(opt_angles_for_graph(G, p))
    # build circuit
    qc = get_maxcut_qaoa_circuit(G, angles["beta"], angles["gamma"])
    qc.measure_all()
    # run circuit
    backend = AerSimulator()
    assert isinstance(backend.run(qc).result().get_counts(), dict)


def test_no_save_state():
    n = 6
    qc = get_maxcut_qaoa_circuit(nx.star_graph(n - 1), [0.1], [0.1], save_state=False)
    creg = ClassicalRegister(n)
    creg_ancilla = ClassicalRegister(1)
    qreg_ancilla = QuantumRegister(1)
    qc.add_register(creg)
    qc.add_register(creg_ancilla)
    qc.add_register(qreg_ancilla)
    qc.h(qreg_ancilla)
    for i in range(n):
        qc.cx(qreg_ancilla, i)
    qc.h(qreg_ancilla)

    qc.measure(range(6), creg)
    qc.measure(qreg_ancilla, creg_ancilla)

    counts = (
        execute(qc, AerSimulator(), basis_gates=["u1", "u2", "u3", "cx"])
        .result()
        .get_counts()
    )
    assert isinstance(counts, dict)


def test_examples_erdos_renyi():
    df = get_20_node_erdos_renyi_graphs()
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 30



import networkx as nx
import copy
from pathlib import Path
import pandas as pd
import numpy as np

from .utils import (
    beta_to_qaoa_format,
    gamma_to_qaoa_format,
    qaoa_maxcut_energy,
)

example_utils_folder = Path(__file__).parent


def get_20_node_erdos_renyi_graphs():
    # First, load the graphs

    n_graphs = 10
    n_qubits = 20

    graphs = {}

    with open(
        Path(
            example_utils_folder,
            "../data/optimal_parameters_n_20/Erdos_Renyi_n_20_density_9pt5_ngraphs_10.txt",
        )
    ) as f:
        for _ in range(1, n_graphs + 1):
            f.readline(-1)  # first line is blank
            line_with_id = f.readline(-1)  # second line has graph number and order
            graph_id = int(line_with_id.strip().split()[1])
            edges = []
            G = nx.Graph()
            for n in range(n_qubits):
                G.add_nodes_from([n])
            # third line is first row of upper triangle of adjacency matrix (without the diagonal element)
            for n in range(n_qubits - 1):
                adj_str = f.readline(-1)
                for m in range(n_qubits - 1 - n):
                    q_num = n + m + 1
                    if adj_str[m] == "1":
                        edges.append([n, q_num])
                        G.add_edge(n, q_num)
            graphs[graph_id] = copy.deepcopy(G)

    graph_id2graph = pd.Series(graphs, name="G")

    # Second, load optimal parameters

    def load_results_file_into_dataframe(p):
        colnames = ["graph_id", "C_{true opt}", "C_init", "C_opt", "pr(max)", "p"]
        for i in range(p):
            colnames.append(f"beta_{i}/pi")
        for i in range(p):
            colnames.append(f"gamma_{i}/pi")
        colnames.append("Dummy 1")
        colnames.append("Dummy 2")
        colnames.append("Dummy 3")

        df = pd.read_csv(
            Path(
                example_utils_folder,
                f"../data/optimal_parameters_n_20/QAOA_dat_p={p}",
            ),
            delim_whitespace=True,
            names=colnames,
            header=None,
            index_col="graph_id",
        )
        df[
            "p_max"
        ] = p  # maximal p allowed; this is to differentiate from p in the original dataset, which can be lower due to achieving optimal solution
        assert (df["p_max"] >= df["p"]).all()
        return df

    df = pd.DataFrame()

    for p in range(1, 4):
        df_orig = load_results_file_into_dataframe(p).merge(
            graph_id2graph, how="outer", right_index=True, left_index=True
        )
        assert len(df_orig) == n_graphs
        df_orig["beta"] = df_orig.apply(
            lambda row: [row[f"beta_{i}/pi"] for i in range(p)], axis=1
        )
        df_orig["gamma"] = df_orig.apply(
            lambda row: [row[f"gamma_{i}/pi"] for i in range(p)], axis=1
        )
        df_orig["theta"] = df_orig.apply(lambda row: row["gamma"] + row["beta"], axis=1)
        df_orig["n"] = df_orig.apply(lambda row: row["G"].number_of_nodes(), axis=1)
        assert (df_orig["n"] == n_qubits).all()
        df = df.append(df_orig.reset_index())
    return df.rename(
        columns={
            "C_opt": "QAOA energy with optimized parameters",
            "C_{true opt}": "Optimal MaxCut solution from brute force",
        }
    )



import numpy as np
import pandas as pd
import networkx as nx
import json
import re
import gc
import copy
import pynauty
import pickle
from pathlib import Path
from io import BytesIO
from urllib.request import urlopen
from zipfile import ZipFile
import shutil
from tqdm import tqdm
from multiprocessing import Process

from .utils import load_results_file_into_dataframe, get_adjacency_dict

build_tables_folder = Path(__file__).parent

n_graphs = {3: 2, 4: 6, 5: 21, 6: 112, 7: 853, 8: 11117, 9: 261080}


def load_data():
    if not Path(build_tables_folder, "../data/qaoa-dataset-version1/").is_dir():
        print("Loading data, this may take a while")
        data_folder = Path(build_tables_folder, "../data/")
        zipurl = "https://github.com/QAOAKit/data/zipball/master"
        with urlopen(zipurl) as zipresp:
            with ZipFile(BytesIO(zipresp.read())) as zfile:
                zfile.extractall(data_folder)
        # remove top folder
        folder = list(Path(build_tables_folder, "../data/").glob("QAOAKit-data-*"))
        assert len(folder) == 1
        folder = folder[0]
        for subfolder in folder.glob("*"):
            if subfolder.stem == ".gitignore":
                continue
            subfolder = subfolder.absolute()
            parent_dir = subfolder.parents[1]
            subfolder.rename(parent_dir / subfolder.name)
        shutil.rmtree(folder)


def build_graph2angles():
    print(f"Building graph2angles table...")

    tables = {}

    for n_qubits in tqdm(range(3, 10)):
        tables[n_qubits] = {}

        for p in tqdm(range(1, 4), leave=False):
            tables[n_qubits][p] = {}

            df = load_results_file_into_dataframe(n_qubits, p)

            for index, data in df.iterrows():
                if data["p"] != p:
                    assert np.isclose(data["C_{true opt}"], data["C_opt"])
                beta = [data[f"beta_{i}/pi"] for i in range(p)]
                gamma = [data[f"gamma_{i}/pi"] for i in range(p)]
                tables[n_qubits][p][int(index)] = {
                    "beta": copy.deepcopy(beta),
                    "gamma": copy.deepcopy(gamma),
                }

    pickle.dump(
        tables,
        open(Path(build_tables_folder, f"../data/lookup_tables/graph2angles.p"), "wb"),
    )
    del tables
    print(f"Done building graph2angles table.")


def build_graph2pynauty():
    print(f"Building graph2pynauty table...")

    tables = {}

    for n_qubits in tqdm(range(3, 10)):

        with open(
            Path(
                build_tables_folder,
                "../data/qaoa-dataset-version1/Graphs/graph" + str(n_qubits) + "c.txt",
            )
        ) as f:
            for graph in tqdm(range(1, n_graphs[n_qubits] + 1), leave=False):
                f.readline(-1)  # first line is blank
                line_with_id = f.readline(-1)  # second line has graph number and order
                graph_id, graph_order = [
                    int(x) for x in re.split(" |, |. |.\n", line_with_id) if x.isdigit()
                ]
                assert graph_order == n_qubits
                edges = []
                G = nx.Graph()
                for n in range(n_qubits):
                    G.add_nodes_from([n])
                # third line is first row of upper triangle of adjacency matrix (without the diagonal element)
                for n in range(n_qubits - 1):
                    adj_str = f.readline(-1)
                    for m in range(n_qubits - 1 - n):
                        q_num = n + m + 1
                        if adj_str[m] == "1":
                            edges.append([n, q_num])
                            G.add_edge(n, q_num)
                g = pynauty.Graph(
                    number_of_vertices=G.number_of_nodes(),
                    directed=nx.is_directed(G),
                    adjacency_dict=get_adjacency_dict(G),
                )
                cert = pynauty.certificate(g)
                tables[cert] = graph_id

    pickle.dump(
        tables,
        open(Path(build_tables_folder, f"../data/lookup_tables/graph2pynauty.p"), "wb"),
    )
    del tables
    print(f"Done building graph2pynauty table.")


def build_graph2pynauty_large():
    print(f"Building graph2pynauty_large tables...")

    for n_qubits in tqdm(range(3, 10)):
        table = {}

        table["graph_id2pynautycert"] = {}
        table["graph_id2graph"] = {}
        table["pynautycert2graph_id"] = {}
        table["pynautycert2graph"] = {}

        with open(
            Path(
                build_tables_folder,
                "../data/qaoa-dataset-version1/Graphs/graph" + str(n_qubits) + "c.txt",
            )
        ) as f:
            for graph in tqdm(range(1, n_graphs[n_qubits] + 1), leave=False):
                f.readline(-1)  # first line is blank
                line_with_id = f.readline(-1)  # second line has graph number and order
                graph_id, graph_order = [
                    int(x) for x in re.split(" |, |. |.\n", line_with_id) if x.isdigit()
                ]
                assert graph_order == n_qubits
                G = nx.Graph()
                edge_id = 0
                for n in range(n_qubits):
                    G.add_nodes_from([n])
                # third line is first row of upper triangle of adjacency matrix (without the diagonal element)
                for n in range(n_qubits - 1):
                    adj_str = f.readline(-1)
                    for m in range(n_qubits - 1 - n):
                        q_num = n + m + 1
                        if adj_str[m] == "1":
                            G.add_edge(n, q_num, edge_id=edge_id)
                            edge_id += 1
                g = pynauty.Graph(
                    number_of_vertices=G.number_of_nodes(),
                    directed=nx.is_directed(G),
                    adjacency_dict=get_adjacency_dict(G),
                )
                cert = pynauty.certificate(g)
                table["graph_id2graph"][graph_id] = copy.deepcopy(G)
                table["graph_id2pynautycert"][graph_id] = cert
                table["pynautycert2graph_id"][cert] = graph_id
                table["pynautycert2graph"][cert] = copy.deepcopy(G)

        assert len(table["graph_id2pynautycert"]) == n_graphs[n_qubits]
        assert len(table["graph_id2graph"]) == n_graphs[n_qubits]
        assert len(table["pynautycert2graph_id"]) == n_graphs[n_qubits]
        assert len(table["pynautycert2graph"]) == n_graphs[n_qubits]
        pickle.dump(
            table,
            open(
                Path(
                    build_tables_folder,
                    f"../data/lookup_tables/graph2pynauty_large_{n_qubits}.p",
                ),
                "wb",
            ),
        )
        del table
        gc.collect()
    print(f"Done building graph2pynauty_large tables.")


def build_full_qaoa_dataset():
    print(f"Building full_qaoa_dataset table...")
    """
    Specifications:
        index: 'pynauty_cert'+p (??)
        columns: [
            'pynauty_cert','graph_id','#nodes','C_{true opt}','C_init','C_opt','pr(max)','p','beta','gamma',
            'theta', # concatenated gamma,beta: theta[idx] = gamma_idx, theta[p+idx] = beta_idx
            'p_max', # maximal p allowed; this is to differentiate from p in the original dataset, which can be lower due to achieving optimal solution
            ]
    """
    df = pd.DataFrame()

    for n_qubits in tqdm(range(3, 10)):
        graph_table = pickle.load(
            open(
                Path(
                    build_tables_folder,
                    f"../data/lookup_tables/graph2pynauty_large_{n_qubits}.p",
                ),
                "rb",
            )
        )
        graph_id2pynauty = pd.Series(
            graph_table["graph_id2pynautycert"], name="pynauty_cert"
        )
        graph_id2graph = pd.Series(graph_table["graph_id2graph"], name="G")
        for p in tqdm(range(1, 4), leave=False):
            df_orig = (
                load_results_file_into_dataframe(n_qubits, p)
                .merge(graph_id2pynauty, how="outer", right_index=True, left_index=True)
                .merge(graph_id2graph, how="outer", right_index=True, left_index=True)
            )
            assert len(df_orig) == n_graphs[n_qubits]
            df_orig["beta"] = df_orig.apply(
                lambda row: [row[f"beta_{i}/pi"] for i in range(p)], axis=1
            )
            df_orig["gamma"] = df_orig.apply(
                lambda row: [row[f"gamma_{i}/pi"] for i in range(p)], axis=1
            )
            df_orig["theta"] = df_orig.apply(
                lambda row: row["gamma"] + row["beta"], axis=1
            )
            df_orig["n"] = df_orig.apply(lambda row: row["G"].number_of_nodes(), axis=1)
            assert (df_orig["n"] == n_qubits).all()
            df = df.append(df_orig.reset_index())
    assert len(df) == sum(v * 3 for v in n_graphs.values())
    df.to_pickle(
        Path(build_tables_folder, "../data/lookup_tables/full_qaoa_dataset_table.p")
    )
    del df
    print(f"Done building full_qaoa_dataset table.")


def build_3_reg_dataset():
    print("Building  3-regular graph table")
    with open(
        Path(build_tables_folder, "../data/3_regular/3r_WURTZ_ensemble.json")
    ) as json_file:
        data = json.load(json_file)

    rows = []
    for row in tqdm(data):
        G = nx.Graph()
        G.add_edges_from(row["edges"])
        g = pynauty.Graph(
            number_of_vertices=G.number_of_nodes(),
            directed=nx.is_directed(G),
            adjacency_dict=get_adjacency_dict(G),
        )
        cert = pynauty.certificate(g)
        c_true_opt = row["0"]["MaxCut"]

        for p in range(1, 11):
            d = {}
            d["G"] = copy.deepcopy(G)
            d["n"] = G.number_of_nodes()
            d["p_max"] = p
            d["pynauty_cert"] = cert
            d["C_{true opt}"] = c_true_opt
            d["C_fixed"] = row[str(p)]["fixed_val"]
            if "optimized_val" in row[str(p)]:
                d["C_opt"] = row[str(p)]["optimized_val"]
                d["beta"] = np.array(row[str(p)]["angles"][0]["beta"]) / np.pi
                d["gamma"] = -2 * np.array(row[str(p)]["angles"][0]["gamma"]) / np.pi
                d["theta"] = np.hstack([d["gamma"], d["beta"]])
            rows.append(copy.deepcopy(d))

    df = pd.DataFrame(rows, columns=rows[0].keys())
    df.to_pickle(
        Path(build_tables_folder, "../data/lookup_tables/3_reg_dataset_table.p")
    )
    print("Done building  3-regular graph table")


if __name__ == "__main__":
    print("Building tables (this may take a few minutes)...")

    # each function is run in its own process to force system to free memory

    p = Process(target=load_data)
    p.start()
    p.join()
    del p
    gc.collect()

    p1 = Process(target=build_graph2angles)
    p1.start()

    p2 = Process(target=build_graph2pynauty)
    p2.start()

    p1.join()
    p2.join()
    del p1
    del p2
    gc.collect()

    p = Process(target=build_graph2pynauty_large)
    p.start()
    p.join()
    del p
    gc.collect()

    p = Process(target=build_full_qaoa_dataset)
    p.start()
    p.join()
    del p
    gc.collect()

    p = Process(target=build_3_reg_dataset)
    p.start()
    p.join()
    del p
    gc.collect()

    print("All done")



import pickle
import copy
import pynauty
import networkx as nx
import numpy as np
import pandas as pd
from pathlib import Path
from functools import partial
from qiskit.providers.aer import AerSimulator
import json

from QAOAKit.qaoa import get_maxcut_qaoa_circuit

utils_folder = Path(__file__).parent


class LookupTableHandler:
    """Singleton handling all the tables
    constructed from the qaoa-dataset-version1

    This object may consume a lot of memory!

    Attributes:
        graph2angles (dict): maps from n_qubits, p and graph_id to optimal parameters
                graph2angles[n_qubits][p][graph_id] = {'beta':optimal_beta, 'gamma':optimal_gamma}
        graph2pynauty (dict): maps from pynauty certificate to graph_id
        full_qaoa_dataset_table (pandas.DataFrame) : TBD
    """

    def __init__(self):
        self.graph2angles = None
        self.graph2pynauty = None
        # dictionary with mapping from nqubits to dictionary containing
        # 'graph_id2graph', 'graph_id2pynautycert', 'pynautycert2graph_id', 'pynautycert2graph' tables
        # example: large_graph_table[5]['graph_id2graph']
        self.large_graph_table = None
        self.full_qaoa_dataset_table = None
        self.three_reg_dataset_table = None
        self.fixed_angle_dataset_table = None

    def get_graph2angles(self):
        if self.graph2angles is None:
            self.graph2angles = pickle.load(
                open(Path(utils_folder, f"../data/lookup_tables/graph2angles.p"), "rb")
            )
        return self.graph2angles

    def get_graph2pynauty(self):
        if self.graph2pynauty is None:
            self.graph2pynauty = pickle.load(
                open(Path(utils_folder, f"../data/lookup_tables/graph2pynauty.p"), "rb")
            )
        return self.graph2pynauty

    def get_large_graph_table(self, nqubits):
        if self.large_graph_table is None:
            self.large_graph_table = {}
        if nqubits not in self.large_graph_table:
            self.large_graph_table[nqubits] = pickle.load(
                open(
                    Path(
                        utils_folder,
                        f"../data/lookup_tables/graph2pynauty_large_{nqubits}.p",
                    ),
                    "rb",
                )
            )
        return self.large_graph_table[nqubits]

    def get_full_qaoa_dataset_table(self):
        if self.full_qaoa_dataset_table is None:
            self.full_qaoa_dataset_table = pd.read_pickle(
                Path(utils_folder, "../data/lookup_tables/full_qaoa_dataset_table.p")
            ).set_index(["pynauty_cert", "p_max"])
        return self.full_qaoa_dataset_table

    def get_3_reg_dataset_table(self):
        if self.three_reg_dataset_table is None:
            self.three_reg_dataset_table = pd.read_pickle(
                Path(utils_folder, "../data/lookup_tables/3_reg_dataset_table.p")
            ).set_index(["pynauty_cert", "p_max"])
        return self.three_reg_dataset_table

    def get_fixed_angle_dataset_table(self):
        # fixed angle dataset is small, no need to store pickle on disk
        if self.fixed_angle_dataset_table is None:
            with open(
                Path(
                    utils_folder,
                    "../data/fixed-angle-2021-08/angles_regular_graphs.json",
                )
            ) as json_file:
                data = json.load(json_file)

            lines = []
            for d in data.keys():
                for p in data[d]:
                    if int(p) < 12:  # remove bad value at p=12
                        angles = [float(x) for x in data[d][p]["angles"]]
                        line_d = {
                            "d": int(d),
                            "p": int(p),
                            "gamma": np.array(angles[: int(p)]) / np.pi,
                            "beta": np.array(angles[int(p) :]) / np.pi,
                            "AR": float(data[d][p]["AR"]),
                        }
                        lines.append(line_d)
            self.fixed_angle_dataset_table = pd.DataFrame(
                lines, columns=lines[0].keys()
            )
        return self.fixed_angle_dataset_table


lookup_table_handler = LookupTableHandler()


def get_adjacency_dict(G):
    """Returns adjacency dictionary for G
    G must be a networkx graph
    Return format: { n : [n1,n2,...], ... }
    where [n1,n2,...] is a list of neighbors of n
    ignores all attributes
    """
    adjacency_dict = {}
    for n, neigh_dict in G.adjacency():
        neigh_list = []
        for neigh, attr_dict in neigh_dict.items():
            neigh_list.append(neigh)
        adjacency_dict[n] = neigh_list
    return adjacency_dict


def isomorphic(G1, G2):
    """Tests if two unweighted graphs are isomorphic using pynauty
    Ignores all attributes
    """
    g1 = pynauty.Graph(
        number_of_vertices=G1.number_of_nodes(),
        directed=nx.is_directed(G1),
        adjacency_dict=get_adjacency_dict(G1),
    )
    g2 = pynauty.Graph(
        number_of_vertices=G2.number_of_nodes(),
        directed=nx.is_directed(G2),
        adjacency_dict=get_adjacency_dict(G2),
    )
    return pynauty.isomorphic(g1, g2)


def get_graph_id(G):
    graph2pynauty = lookup_table_handler.get_graph2pynauty()

    g = pynauty.Graph(
        number_of_vertices=G.number_of_nodes(),
        directed=nx.is_directed(G),
        adjacency_dict=get_adjacency_dict(G),
    )
    cert = pynauty.certificate(g)

    return graph2pynauty[cert]


def get_graph_from_id(graph_id, nqubits):
    graph_id2graph = lookup_table_handler.get_large_graph_table(nqubits)[
        "graph_id2graph"
    ]
    return copy.deepcopy(graph_id2graph[graph_id])


def opt_angles_for_graph(G, p):
    if (G.number_of_nodes() <= 9) and (p <= 3):
        graph2angles = lookup_table_handler.get_graph2angles()
        graph_id = get_graph_id(G)
        return copy.deepcopy(graph2angles[G.number_of_nodes()][p][graph_id])
    elif nx.is_regular(G) and G.degree[0] == 3 and p <= 2:
        row = get_3_reg_dataset_table_row(G, p)
        return {"beta": row["beta"], "gamma": row["gamma"]}
    else:
        raise NotImplementedError(
            "Should return angles from fixed angle conjecture paper; TBD"
        )


def get_fixed_angles(d, p):
    df = lookup_table_handler.get_fixed_angle_dataset_table()
    row = df[(df["d"] == d) & (df["p"] == p)]
    assert len(row) == 1
    row = row.squeeze()
    assert isinstance(row, pd.Series)
    angles = {"beta": row.beta, "gamma": row.gamma}
    return angles


def get_full_qaoa_dataset_table():
    return lookup_table_handler.get_full_qaoa_dataset_table()


def get_3_reg_dataset_table():
    return lookup_table_handler.get_3_reg_dataset_table()


def get_full_qaoa_dataset_table_row(G, p):
    """Returns full table row for a given NetworkX graph"""
    full_qaoa_dataset_table = lookup_table_handler.get_full_qaoa_dataset_table()

    g = pynauty.Graph(
        number_of_vertices=G.number_of_nodes(),
        directed=nx.is_directed(G),
        adjacency_dict=get_adjacency_dict(G),
    )
    cert = pynauty.certificate(g)

    return full_qaoa_dataset_table.loc[(cert, p)]


def get_3_reg_dataset_table_row(G, p):
    """Returns full table row for a given NetworkX graph"""
    df = lookup_table_handler.get_3_reg_dataset_table()

    g = pynauty.Graph(
        number_of_vertices=G.number_of_nodes(),
        directed=nx.is_directed(G),
        adjacency_dict=get_adjacency_dict(G),
    )
    cert = pynauty.certificate(g)

    return df.loc[(cert, p)]


def angles_to_qaoa_format(angles):
    """Converts from format in graph2angles
    into the format used by qaoa.py
    get_maxcut_qaoa_circuit(G, angles['beta'], angles['gamma'])
    """
    res = copy.deepcopy(angles)
    res["beta"] = beta_to_qaoa_format(res["beta"])
    res["gamma"] = gamma_to_qaoa_format(res["gamma"])
    return res


def beta_to_qaoa_format(beta):
    """Converts from format in graph2angles
    into the format used by qaoa.py
    """
    return np.pi * np.array(beta)


def gamma_to_qaoa_format(gamma):
    """Converts from format in graph2angles
    into the format used by qaoa.py
    get_maxcut_qaoa_circuit(G, angles['beta'], angles['gamma'])
    """
    return -np.pi * np.array(gamma) / 2


def angles_to_qiskit_format(angles):
    """Converts from format in graph2angles
    into the format used by QAOAAnsatz
    """
    return np.concatenate(
        [[-np.pi * g, np.pi * b] for g, b in zip(angles["gamma"], angles["beta"])]
    )


def angles_from_qiskit_format(angles):
    """Converts from the format used by QAOAAnsatz
    into the format in graph2angles
    """
    res = {}
    assert len(angles) % 2 == 0
    res["gamma"] = list(x / (-np.pi) for x in angles[::2])
    res["beta"] = list(x / np.pi for x in angles[1::2])
    return res


def angles_to_qtensor_format(angles):
    """Converts from format in graph2angles
    into the format used by QTensor
    """
    return {"gamma": [-g / 2 for g in angles["gamma"]], "beta": angles["beta"]}


def load_results_file_into_dataframe(n_qubits, p):
    """Loads one file from ../data/qaoa-dataset-version1/Results/ into a pandas.DataFrame
    Column names are from ../data/qaoa-dataset-version1/Results/How_to_read_data_columns.txt
    One column is added:
    p_max : maximal p allowed; this is to differentiate from p in the original dataset, which can be lower due to achieving optimal solution
    """
    colnames = ["graph_id", "C_{true opt}", "C_init", "C_opt", "pr(max)", "p"]
    for i in range(p):
        colnames.append(f"beta_{i}/pi")
    for i in range(p):
        colnames.append(f"gamma_{i}/pi")
    df = pd.read_csv(
        Path(
            utils_folder,
            f"../data/qaoa-dataset-version1/Results/p={p}/n={n_qubits}_p={p}.txt",
        ),
        delim_whitespace=True,
        names=colnames,
        header=None,
        index_col="graph_id",
    )
    df[
        "p_max"
    ] = p  # maximal p allowed; this is to differentiate from p in the original dataset, which can be lower due to achieving optimal solution
    assert (df["p_max"] >= df["p"]).all()
    return df


def get_graph_and_assign_weights(graph_id, weight_id, nqubits, df_weights):
    """Retrieves a graph from graph_id and nqubits and assigns weights
    from df_weights
    """
    weights = list(
        df_weights[
            (df_weights["graph_id"] == graph_id)
            & (df_weights["weight_id"] == weight_id)
        ]["weights"]
    )
    assert len(weights) == 1
    weights = weights[0]
    G = get_graph_from_id(graph_id, nqubits)
    for u, v, attr_dict in G.edges(data=True):
        G[u][v]["weight"] = weights[attr_dict["edge_id"]]
    return copy.deepcopy(G)


def load_weighted_results_into_dataframe(folder_path, p, nqubits, df_weights):
    """Loads all result files from ../data/weighted_angle_dat/{p}
    df_weights is a dataframe mapping graph_id and weight_id to list of weights
    The column names and conventions are described in ../data/weighted_angle_dat/Readme.txt
    One column is added:
    p_max : maximal p allowed; this is to differentiate from p in the original dataset, which can be lower due to achieving optimal solution
    """
    colnames = [
        "graph_id",
        "weight_id",
        "C_{true opt}",
        "C_init",
        "C_opt",
        "pr(max)",
        "p",
    ]
    for i in range(p):
        colnames.append(f"beta_{i}/pi")
    for i in range(p):
        colnames.append(f"gamma_{i}/pi")
    colnames.append("mean(weight)")
    colnames.append("std(weight)")
    dfs = []
    for fname in folder_path.glob("QAOA_dat_weighted_*"):
        dfs.append(
            pd.read_csv(fname, delim_whitespace=True, names=colnames, header=None)
        )
    df = pd.concat(dfs)
    df[
        "p_max"
    ] = p  # maximal p allowed; this is to differentiate from p in the original dataset, which can be lower due to achieving optimal solution
    df["beta"] = df.apply(lambda row: [row[f"beta_{i}/pi"] for i in range(p)], axis=1)
    df["gamma"] = df.apply(lambda row: [row[f"gamma_{i}/pi"] for i in range(p)], axis=1)
    df["theta"] = df.apply(lambda row: row["gamma"] + row["beta"], axis=1)
    df["G"] = df.apply(
        lambda row: get_graph_and_assign_weights(
            row["graph_id"], row["weight_id"], nqubits, df_weights
        ),
        axis=1,
    )
    assert np.all(
        np.isclose(
            df.apply(
                lambda row: np.mean(
                    [x[2]["weight"] for x in row["G"].edges(data=True)]
                ),
                axis=1,
            ),
            df["mean(weight)"],
            atol=1e-07,
        )
        | np.isnan(df["std(weight)"])
    )

    assert np.all(
        np.isclose(
            df.apply(
                lambda row: np.std([x[2]["weight"] for x in row["G"].edges(data=True)]),
                axis=1,
            ),
            df["std(weight)"],
            atol=1e-07,
        )
        | np.isnan(df["std(weight)"])
    )

    assert np.all(
        np.isclose(
            df.head(100).apply(
                lambda row: qaoa_maxcut_energy(
                    row["G"],
                    beta_to_qaoa_format(row["beta"]),
                    gamma_to_qaoa_format(row["gamma"]),
                ),
                axis=1,
            ),
            df.head(100)["C_opt"],
        )
    )

    return df


def load_weights_into_dataframe(folder_path):
    """Loads all weight files from the folder passed
    the first column is the graph number
    The next E columns are the weights of the E edges.
    Other columns beyond E+1 are meaningless
    These weights were drawn uniformly at random from 1,2
    returns pandas.DataFrame with columns
       -'weight_id' (int)
       -'graph_id' (int)
       -'weights' (list of floating point weights)
    """
    lines = []

    for fname in folder_path.glob("weights_*"):
        with open(fname, "r") as f:
            for line_num, line in enumerate(f.readlines()):
                line = line.strip().split()
                line_d = {
                    "weight_id": line_num + 1,
                    "graph_id": int(line[0]),
                    "weights": [float(x) for x in line[1:]],
                }
                lines.append(line_d)
    return pd.DataFrame(lines, columns=lines[0].keys())


#############################
# QAOA utils
############################


def state_num2str(basis_state_as_num, nqubits):
    return "{0:b}".format(basis_state_as_num).zfill(nqubits)


def state_str2num(basis_state_as_str):
    return int(basis_state_as_str, 2)


def state_reverse(basis_state_as_num, nqubits):
    basis_state_as_str = state_num2str(basis_state_as_num, nqubits)
    new_str = basis_state_as_str[::-1]
    return state_str2num(new_str)


def get_adjusted_state(state):
    nqubits = np.log2(state.shape[0])
    if nqubits % 1:
        raise ValueError("Input vector is not a valid statevector for qubits.")
    nqubits = int(nqubits)

    adjusted_state = np.zeros(2 ** nqubits, dtype=complex)
    for basis_state in range(2 ** nqubits):
        adjusted_state[state_reverse(basis_state, nqubits)] = state[basis_state]
    return adjusted_state


def state_to_ampl_counts(vec, eps=1e-15):
    """Converts a statevector to a dictionary
    of bitstrings and corresponding amplitudes
    """
    qubit_dims = np.log2(vec.shape[0])
    if qubit_dims % 1:
        raise ValueError("Input vector is not a valid statevector for qubits.")
    qubit_dims = int(qubit_dims)
    counts = {}
    str_format = "0{}b".format(qubit_dims)
    for kk in range(vec.shape[0]):
        val = vec[kk]
        if val.real ** 2 + val.imag ** 2 > eps:
            counts[format(kk, str_format)] = val
    return counts


def obj_from_statevector(sv, obj_f):
    """Compute objective from Qiskit statevector
    For large number of qubits, this is slow.
    """
    qubit_dims = np.log2(sv.shape[0])
    if qubit_dims % 1:
        raise ValueError("Input vector is not a valid statevector for qubits.")
    qubit_dims = int(qubit_dims)
    # get bit strings for each element of the state vector
    # https://stackoverflow.com/questions/22227595/convert-integer-to-binary-array-with-suitable-padding
    bit_strings = (
        ((np.array(range(sv.shape[0]))[:, None] & (1 << np.arange(qubit_dims)))) > 0
    ).astype(int)

    return sum(
        obj_f(bit_strings[kk]) * (np.abs(sv[kk]) ** 2) for kk in range(sv.shape[0])
    )


def maxcut_obj(x, w):
    """Compute the value of a cut.

    Args:
        x (numpy.ndarray): binary string as numpy array.
        w (numpy.ndarray): adjacency matrix.

    Returns:
        float: value of the cut.
    """
    X = np.outer(x, (1 - x))
    return np.sum(w * X)


def get_adjacency_matrix(G):
    n = G.number_of_nodes()
    w = np.zeros([n, n])

    for e in G.edges():
        if nx.is_weighted(G):
            w[e[0], e[1]] = G[e[0]][e[1]]["weight"]
            w[e[1], e[0]] = G[e[0]][e[1]]["weight"]
        else:
            w[e[0], e[1]] = 1
            w[e[1], e[0]] = 1
    return w


def qaoa_maxcut_energy(G, beta, gamma):
    """Computes MaxCut QAOA energy for graph G
    qaoa format (`angles_to_qaoa_format`) used for beta, gamma
    """
    obj = partial(maxcut_obj, w=get_adjacency_matrix(G))
    qc = get_maxcut_qaoa_circuit(G, beta, gamma)
    backend = AerSimulator(method="statevector")
    sv = backend.run(qc).result().get_statevector()
    return obj_from_statevector(sv, obj)



# QAOA circuit for MAXCUT

import networkx as nx
from qiskit import QuantumCircuit, Aer, execute
from qiskit.compiler import transpile
from qiskit_optimization import QuadraticProgram
from qiskit.algorithms.minimum_eigen_solvers.qaoa import QAOAAnsatz
import numpy as np


def append_zz_term(qc, q1, q2, gamma):
    qc.cx(q1, q2)
    qc.rz(2 * gamma, q2)
    qc.cx(q1, q2)


def get_maxcut_cost_operator_circuit(G, gamma):
    N = G.number_of_nodes()
    qc = QuantumCircuit(N)
    for i, j in G.edges():
        if nx.is_weighted(G):
            append_zz_term(qc, i, j, gamma * G[i][j]["weight"])
        else:
            append_zz_term(qc, i, j, gamma)
    return qc


def append_x_term(qc, q1, beta):
    qc.h(q1)
    qc.rz(2 * beta, q1)
    qc.h(q1)


def get_mixer_operator_circuit(G, beta):
    N = G.number_of_nodes()
    qc = QuantumCircuit(N)
    for n in G.nodes():
        append_x_term(qc, n, beta)
    return qc


def get_maxcut_qaoa_circuit(G, beta, gamma, transpile_to_basis=True, save_state=True):
    assert len(beta) == len(gamma)
    p = len(beta)  # infering number of QAOA steps from the parameters passed
    N = G.number_of_nodes()
    qc = QuantumCircuit(N)
    # first, apply a layer of Hadamards
    qc.h(range(N))
    # second, apply p alternating operators
    for i in range(p):
        qc = qc.compose(get_maxcut_cost_operator_circuit(G, gamma[i]))
        qc = qc.compose(get_mixer_operator_circuit(G, beta[i]))
    if transpile_to_basis:
        qc = transpile(qc, optimization_level=0, basis_gates=["u1", "u2", "u3", "cx"])
    if save_state:
        qc.save_state()
    return qc


def get_maxcut_qaoa_qiskit_circuit(G, p, qiskit_angles):
    """
    Constructs a max cut QAOA circuit with Qiskit.

    Returns (quantum circuit, cost operator, cut size offset).
    On average, the circuit gives max cut size proportional to -(<C> + offset)
    """
    n_qubits = len(G.nodes())
    problem = QuadraticProgram()
    _ = [problem.binary_var("x{}".format(i)) for i in range(n_qubits)]
    problem.maximize(
        linear=nx.adjacency_matrix(G).dot(np.ones(n_qubits)),
        quadratic=-nx.adjacency_matrix(G),
    )
    C, offset = problem.to_ising()
    ansatz = QAOAAnsatz(C, p).decompose()
    qc = ansatz.bind_parameters(qiskit_angles)
    qc.save_state()
    return qc, C, offset



from setuptools import setup

setup(
    name="QAOAKit",
    version="0.1.7",
    description="A Toolkit for Reproducible Study, Application and Verification of QAOA",
    author="Ruslan Shaydulin",
    author_email="ruslan@shaydul.in",
    python_requires=">=3, <4",
    packages=["QAOAKit"],
    install_requires=[
        "qiskit>=0.28.0",
        "pynauty==1.0.0",
        "qiskit-optimization",
        "pandas",
        "networkx",
        "numpy",
        "pytest",
        "tqdm",
    ],
    zip_safe=True,
)
